import events
from pygame.sprite import Sprite
from pygame import Rect
import pygame
from utils import *
import math
import random
from copy import copy

DRAGTWIST_RATIO = 80
DRAGSTRETCH_RATIO = 20

def IsOutOfBounds( sprite ):
	r = sprite.rect
	return ( r.left  > 800 or r.right < 0 \
	      or r.top > 600 or r.bottom < 0 )

def ChangeAndDirtyAttr( attrName ):
	def Setter(self, val):
		#print 'setting ', attrName, val
		if val == self.__dict__[attrName]:
			return
		self.__dict__[attrName] = val
		self._imageDirty = True
	def Getter(self):
		return self.__dict__[attrName]

	return property(Getter, Setter)

class Focusable:
	def __init__(self):
		self._focused = False
		self._imageDirty = True
	focused = ChangeAndDirtyAttr('_focused')
class Dragable:
	def __init__(self):
		self._dragging = False
		self.dragStartPoint = None
		self._imageDirty = True
	dragging = ChangeAndDirtyAttr('_dragging')

class Dot(Sprite):
	red = 0
	def __init__(self, radius):
		Sprite.__init__(self)
		self.radius = radius
		self.image = TransSurface( self.radius*2, self.radius*2 )
		if DEBUG:
			color = (Dot.red%256,Dot.red%256,255,255)
			Dot.red += 10
		else:
			color = (255,255,255,255)
		pygame.draw.circle( self.image, color, 
		                    (self.radius,self.radius), self.radius )
		self.rect = self.image.get_rect()

	def ChangeColor( self, color ):
		Dot.red = (Dot.red + 22) %256
		pct = float(Dot.red)/255
		color = (((pct* color[0] )+20) %256,
		         ((pct* color[1] )+20) %256,
		         ((pct* color[2] )+20) %256,
		         255)
		print color
		pygame.draw.circle( self.image, color, 
		                    (self.radius,self.radius), self.radius )



class ElasticBand(Sprite, Focusable, Dragable):
	def __init__(self, parent=None, radius=6, width=20):
		Sprite.__init__(self)
		Focusable.__init__(self)
		Dragable.__init__(self)
		events.AddListener(self)

		self.parent = parent
		self.radius = radius
		self.width = width
		self.angle = 0

		self.stretch = 40
		self.stretchFloor = 40
		self.stretchCeiling = 200

		self.twist = 0
		self.twistFloor = 0
		self.twistCeiling = 3

		self.dots = [
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		             Dot(self.radius),
		            ]

		self.active = True

		self._cachedImage = None

		self._originPoint = [0,0]
		self.CalcDotPositions()

	def Die( self ):
		events.Fire('GamePieceDeath', self )
	

	
	def SetOriginPoint(self, pos):
		self._originPoint = pos
		self.CalcDotPositions()

	def GetOriginPoint( self ):
		#print 'someone is getting o point'
		return self._originPoint
	
	originPoint = property( GetOriginPoint, SetOriginPoint )

	def ChangeColor( self, color ):
		for dot in self.dots:
			dot.ChangeColor( color )


	def GetImage(self):
		if not self._imageDirty:
			return self._cachedImage
		self.CalcDotPositions()
		newImg = TransSurface( abs(self.rect.width), 
		                       abs(self.rect.height) )
		for dot in ds:
			newImg.blit( dot.image, dot.rect )
		self._cachedImage = newImg
		self._imageDirty = False
		return newImg

	def SetImage(self, value):
		print "oh really fool?"

	#image = property(GetImage, SetImage)

	def CalcDotPositions(self):
		ds = self.dots
		oPt = self.originPoint
		minX, minY = oPt
		maxX, maxY = oPt
		pairs = zip( ds[::2], ds[1::2] )
		stretchStep = self.stretch/len(pairs)
		twistStep = self.twist/len(pairs)

		for i, pair in enumerate(pairs):
			pair[0].rect.center = oPt[0], oPt[1] - self.width/2
			pair[1].rect.center = oPt[0], oPt[1] + self.width/2

			pair[0].rect.centerx += i*stretchStep
			pair[1].rect.centerx += i*stretchStep

			thisTwist = math.sin(math.pi*twistStep*i)*self.width
			thisTwist = abs(thisTwist)
			pair[0].rect.centery += thisTwist
			pair[1].rect.centery -= thisTwist

			for dot in pair:
				minX = min( dot.rect.left, minX )
				maxX = max( dot.rect.right, maxX )
				minY = min( dot.rect.top, minY )
				maxY = max( dot.rect.bottom, maxY )
		self.rect = Rect( minX, minY, (maxX-minX), (maxY-minY) )


	def StretchMore( self, furtherStretch ):
		self.stretch += furtherStretch/DRAGSTRETCH_RATIO
		self.stretch = Clamp( self.stretch, 
		                      self.stretchFloor, self.stretchCeiling )
		twistAmt = (self.stretch - self.stretchFloor) /\
		           (self.stretchCeiling - self.stretchFloor)
		self.twist = (self.twistCeiling - self.twistFloor) * twistAmt
	
	def SetAngle(self, guideAngle):
		changeAngle = guideAngle - self.angle
		self.angle = guideAngle
		originPt = self.originPoint

		for dot in self.dots:
			dCenter = dot.rect.center
			distance = Distance( originPt, dCenter )
			origAngle = GetAngle( dCenter, originPt )
			newAngle = guideAngle + origAngle
			newPos = GetPosAtAngleFrom(originPt, newAngle, distance)

			dot.rect.center = newPos
		
	def update(self, timeChange=1):
		if not self.dragging:
			self.update_UnwindStep(timeChange)
		if not self._imageDirty:
			return
		self.CalcDotPositions()
	
	def update_UnwindStep(self, timeChange):
		stretchStrength = self.stretch - self.stretchFloor
		if self.dragging or stretchStrength < 1:
			return
		maxStrength = self.stretchCeiling - self.stretchFloor
		power = 60.0 * float(stretchStrength)/maxStrength
		self.StretchMore( -power )
		self.parent.Wind( power )
		self._imageDirty = True

	def On_MouseDown(self, event):
		if not self.rect.collidepoint( event.pos ):
			self.focused = False
			return

		self.focused = True
		self.dragging = True
		self.dragStartPoint = event.pos

	def On_MouseUp(self, event):
		return
		#print 'on mouse up'
		#if not self.focused:
			#return
#
		#self.dragging = False
		
	def On_MouseMotion(self, event):
		return
		#print 'handling mousemouion'
		#if not self.focused and self.rect.collidepoint( event.pos ):
		#	self.focused = True
		#if not self.dragging:
		#	return

		#self._imageDirty = True
		#currentDragDistance = Distance( event.pos, self.dragStartPoint )
		#originToDragStart = Distance( self.originPoint, self.dragStartPoint )
		#originToDragPos = Distance( self.originPoint, event.pos )
		#if originToDragPos < originToDragStart:
		#	#reverse it if dragging towards the origin
		#	currentDragDistance *= -1
		#self.StretchMore( currentDragDistance )

			
		
class MarbleFactory(Sprite):
	timeToLaunch = 400
	def __init__(self):
		Sprite.__init__(self)
		events.AddListener( self )
		self.radius = 30
		self.image = TransSurface( self.radius*2, self.radius*2 )
		color = (0,0,0,255)
		pygame.draw.circle( self.image, color, 
		                    (self.radius,self.radius), self.radius )

		self.rect = self.image.get_rect()
		self.rect.center = [100,400]
		self.lastLaunch = 0
		self.halted = False
		self.chamberedMarble = None
	
	def StartProduction(self):
		self.halted = False
	def StopProduction(self):
		self.halted = True

	def update( self, timeChange ):
		if self.chamberedMarble:
			self.lastLaunch = 0
		else:
			self.lastLaunch += timeChange
		if not self.halted and self.lastLaunch > self.timeToLaunch:
			elasticBand = ElasticBand()
			self.chamberedMarble = Marble( elasticBand )
			self.chamberedMarble.SetCenter( self.rect.center )
			events.Fire("GamePieceBirth", elasticBand)
			events.Fire("AvatarBirth", self.chamberedMarble)

	def On_MarbleSnapped( self, marble ):
		if marble == self.chamberedMarble:
			self.chamberedMarble = None
			self.lastLaunch = 0

class Marble(Sprite, Focusable, Dragable):
	def __init__(self, elasticBand):
		Sprite.__init__(self)
		Focusable.__init__(self)
		Dragable.__init__(self)
		self.elasticBand = elasticBand
		self.elasticBand.parent = self
		self.angle = 0
		self.speed = 0
		events.AddListener(self)
		self._cachedAngle = 0

		self.dragStartPoint = None
		self.dragEndPoint = None

		self.radius = 20
		self.color = (255,255,255,255)
		self.MakeImage()
		self.rect = self.image.get_rect()
		self._floatCenter = self.rect.center

	def MakeImage( self ):
		self.image = TransSurface( self.radius*2, self.radius*2 )
		pygame.draw.circle( self.image, self.color, 
		                    (self.radius,self.radius), self.radius )

	def SetCenter( self, center ):
		self._floatCenter = center
		self.rect.center = self._floatCenter
		self.elasticBand.SetOriginPoint( self._floatCenter )


	def Die( self ):
		self.elasticBand.Die()
		events.Fire('AvatarDeath', self )
	
	def GetVelocity(self):
		if self.speed < 0.5:
			return (0,0)
		return GetPosAtAngleFrom( (0,0), self.angle, self.speed )

	velocity = property(GetVelocity)

	def update( self, timeChange ):
		if self.dragging:
			self.Drag()
		self.elasticBand.SetOriginPoint( self._floatCenter )
		self.elasticBand.update( timeChange )
		self.update_AngleDots()
		self._floatCenter = VectorAdd(self._floatCenter, self.velocity)
		movement = VectorSubtract(self._floatCenter, self.rect.center)
		self.rect.move_ip( *movement )

		if IsOutOfBounds( self ):
			self.Die()

		self.update_CollideColorChange()

	def update_CollideColorChange(self):
		if self.speed < 0.5:
			return
		collider = self.GetCollider()
		if collider:
			self.angle = GetAngle(
			                       self._floatCenter,
			                       collider._floatCenter
			                     )
			if collider.color == (255,255,255,255):
				self.color = (random.randint(0,255),
				              random.randint(0,255),
				              random.randint(0,255),
				              255 )
				self.elasticBand.ChangeColor( self.color )
				collider.color = copy(self.color)
				collider.elasticBand.ChangeColor( self.color )
			else:
				newCol = list(self.color)
				idx = random.randint( 0,2 )
				newCol[idx] = random.choice(collider.color[:3])
				self.color = newCol
				self.elasticBand.ChangeColor( newCol )
			self.MakeImage()
			collider.MakeImage()

	def GetCollider( self ):
		myGroup = self.groups()[0] #TODO XXX BIG HACK
		colliders = pygame.sprite.spritecollide( self, myGroup, False )
		for c in colliders:
			if c != self:
				return c
		
	def update_AngleDots(self):
		if self.dragging:
			guideAngle = GetAngle( self.dragEndPoint, 
			                       self.rect.center )
		else:
			oppositeOfMoveAngle = self.angle - math.pi
			guideAngle = oppositeOfMoveAngle
		guideAngle = NormalizeAngle( guideAngle )
		#if round(guideAngle,3) == round(self._cachedAngle,3):
			#return
		self._cachedAngle = guideAngle
		#print '--------- angle', guideAngle, guideAngle*180/math.pi
		self.elasticBand.SetAngle( guideAngle )

	def Wind( self, power ):
		self.speed = power/5

	def On_MouseDown(self, event):
		if not self.rect.collidepoint( event.pos ):
			self.focused = False
			return

		self.focused = True
		self.dragging = True
		self.elasticBand.dragging = True
		self.dragStartPoint = event.pos
		self.dragEndPoint = event.pos

	def On_MouseUp(self, event):
		if not self.focused:
			return

		if self.dragging:
			dragAngle = GetAngle( self.dragEndPoint, 
			                      self.rect.center )
			oppositeOfDragAngle = dragAngle - math.pi
			self.angle = oppositeOfDragAngle

		self.dragging = False
		self.elasticBand.dragging = False
		events.Fire( 'MarbleSnapped', self )


	def On_MouseMotion(self, event):
		if not self.focused and self.rect.collidepoint( event.pos ):
			self.focused = True
		if not self.dragging:
			return

		self._imageDirty = True
		self.dragEndPoint = event.pos

	def Drag(self):
		currentDragDistance = Distance( self.dragEndPoint, 
		                                self.dragStartPoint )
		centerToDragPos = Distance( self.rect.center, 
		                            self.dragEndPoint )
		centerToDragStart = Distance( self.rect.center, 
		                              self.dragStartPoint )
		if centerToDragPos < centerToDragStart:
			#reverse it if dragging towards the center
			currentDragDistance *= -1
			stretchAmount = currentDragDistance
		else:
			stretchAmount = centerToDragStart + currentDragDistance
		self.elasticBand.StretchMore( stretchAmount )

